from __future__ import annotations

from dataclasses import dataclass
from functools import partial, wraps
from types import (
    BuiltinFunctionType,
    FunctionType,
    MethodDescriptorType,
    MethodType,
    MethodWrapperType,
    WrapperDescriptorType,
)
from typing import TYPE_CHECKING, Any, TypeVar, overload

from typing_extensions import ParamSpec, override

if TYPE_CHECKING:
    from collections.abc import Callable, Generator

_P = ParamSpec("_P")
_T = TypeVar("_T")
_U = TypeVar("_U")
_V = TypeVar("_V")


def ensure_not_none(obj: _T | None, /) -> _T:
    """Ensure an object is not None."""
    if obj is None:
        raise EnsureNotNoneError(obj=obj)
    return obj


@dataclass(kw_only=True)
class EnsureNotNoneError(Exception):
    obj: Any

    @override
    def __str__(self) -> str:
        return f"Object {self.obj} must not be None"


def first(pair: tuple[_T, Any], /) -> _T:
    """Get the first element in a pair."""
    return pair[0]


@overload
def get_class(obj: type[_T], /) -> type[_T]: ...
@overload
def get_class(obj: _T, /) -> type[_T]: ...
def get_class(obj: _T | type[_T], /) -> type[_T]:
    """Get the class of an object, unless it is already a class."""
    return obj if isinstance(obj, type) else type(obj)


def get_class_name(obj: Any, /) -> str:
    """Get the name of the class of an object, unless it is already a class."""
    return get_class(obj).__name__


def get_func_name(obj: Callable[..., Any], /) -> str:
    """Get the name of a callable."""
    if isinstance(obj, BuiltinFunctionType | FunctionType | MethodType):
        return obj.__name__
    if isinstance(
        obj, MethodDescriptorType | MethodWrapperType | WrapperDescriptorType
    ):
        return obj.__qualname__
    if isinstance(obj, partial):
        return get_func_name(obj.func)
    return get_class_name(obj)


def identity(obj: _T, /) -> _T:
    """Return the object itself."""
    return obj


def is_none(obj: Any, /) -> bool:
    """Check if an object is `None`."""
    return obj is None


def is_not_none(obj: Any, /) -> bool:
    """Check if an object is not `None`."""
    return obj is not None


def not_func(func: Callable[_P, bool], /) -> Callable[_P, bool]:
    """Lift a boolean-valued function to return its conjugation."""

    @wraps(func)
    def wrapped(*args: _P.args, **kwargs: _P.kwargs) -> bool:
        return not func(*args, **kwargs)

    return wrapped


def second(pair: tuple[Any, _U], /) -> _U:
    """Get the second element in a pair."""
    return pair[1]


def send_and_next(value: _U, generator: Generator[_T | None, _U | None, Any], /) -> _T:
    """Send a value to a generator, and then yield the output."""
    result = ensure_not_none(generator.send(value))
    _ = next(generator)
    return result


def start_generator_coroutine(
    func: Callable[_P, Generator[_T, _U, _V]], /
) -> Callable[_P, Generator[_T, _U, _V]]:
    """Instantiate and then start a generator-coroutine."""

    @wraps(func)
    def wrapped(*args: _P.args, **kwargs: _P.kwargs) -> Generator[_T, _U, _V]:
        coro = func(*args, **kwargs)
        _ = next(coro)
        return coro

    return wrapped


__all__ = [
    "EnsureNotNoneError",
    "ensure_not_none",
    "first",
    "get_class",
    "get_class_name",
    "get_func_name",
    "identity",
    "is_none",
    "is_not_none",
    "not_func",
    "second",
    "send_and_next",
    "start_generator_coroutine",
]



from __future__ import annotations

from math import inf, nan
from typing import TYPE_CHECKING, ClassVar, cast

import redis
import redis.asyncio
from hypothesis import assume, given
from hypothesis.strategies import (
    DataObject,
    SearchStrategy,
    data,
    floats,
    sampled_from,
    tuples,
)
from polars import Boolean, DataFrame, DataType, Float64, Int64, Utf8
from polars.testing import assert_frame_equal
from pytest import raises
from redis.commands.timeseries import TimeSeries

from tests.conftest import SKIPIF_CI_AND_NOT_LINUX
from utilities.datetime import EPOCH_UTC, drop_microseconds
from utilities.hypothesis import (
    int32s,
    lists_fixed_length,
    redis_cms,
    text_ascii,
    zoned_datetimes,
)
from utilities.polars import DatetimeUTC, check_polars_dataframe, zoned_datetime
from utilities.redis import (
    _TimeSeriesAddDataFrameKeyIsNotUtf8Error,
    _TimeSeriesAddDataFrameKeyMissingError,
    _TimeSeriesAddDataFrameTimestampIsNotAZonedDatetimeError,
    _TimeSeriesAddDataFrameTimestampMissingError,
    _TimeSeriesAddErrorAtUpsertError,
    _TimeSeriesAddInvalidTimestampError,
    _TimeSeriesAddInvalidValueError,
    _TimeSeriesMAddInvalidKeyError,
    _TimeSeriesMAddInvalidTimestampError,
    _TimeSeriesMAddInvalidValueError,
    _TimeSeriesMAddKeyIsNotUtf8Error,
    _TimeSeriesMAddKeyMissingError,
    _TimeSeriesMAddTimestampIsNotAZonedDatetimeError,
    _TimeSeriesMAddTimestampMissingError,
    _TimeSeriesMAddValueIsNotNumericError,
    _TimeSeriesMAddValueMissingError,
    _TimeSeriesRangeInvalidKeyError,
    _TimeSeriesRangeKeyWithInt64AndFloat64Error,
    _TimeSeriesRangeNoKeysRequestedError,
    _TimeSeriesReadDataFrameNoColumnsRequestedError,
    _TimeSeriesReadDataFrameNoKeysRequestedError,
    ensure_time_series_created,
    ensure_time_series_created_async,
    time_series_add,
    time_series_add_async,
    time_series_add_dataframe,
    time_series_add_dataframe_async,
    time_series_get,
    time_series_get_async,
    time_series_madd,
    time_series_madd_async,
    time_series_range,
    time_series_range_async,
    time_series_read_dataframe,
    time_series_read_dataframe_async,
    yield_client,
    yield_client_async,
    yield_time_series,
    yield_time_series_async,
)
from utilities.zoneinfo import UTC, HongKong

if TYPE_CHECKING:
    import datetime as dt
    from zoneinfo import ZoneInfo

    from polars._typing import SchemaDict

    from utilities.types import Number


valid_zoned_datetimes = zoned_datetimes(
    min_value=EPOCH_UTC, time_zone=sampled_from([HongKong, UTC]), valid=True
).map(drop_microseconds)
invalid_zoned_datetimes = (
    zoned_datetimes(
        max_value=EPOCH_UTC, time_zone=sampled_from([HongKong, UTC]), valid=True
    )
    .map(drop_microseconds)
    .filter(lambda t: t < EPOCH_UTC)
)


@SKIPIF_CI_AND_NOT_LINUX
class TestEnsureTimeSeriesCreated:
    @given(data=data())
    async def test_main(self, *, data: DataObject) -> None:
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis() as client:
                    assert client.exists(container.key) == 0
                    for _ in range(2):
                        ensure_time_series_created(client.ts(), container.key)
                    assert client.exists(container.key) == 1
                case redis.asyncio.Redis() as client:
                    assert await client.exists(container.key) == 0
                    for _ in range(2):
                        await ensure_time_series_created_async(
                            client.ts(), container.key
                        )
                    assert await client.exists(container.key) == 1


@SKIPIF_CI_AND_NOT_LINUX
class TestTimeSeriesAddAndGet:
    @given(
        data=data(),
        timestamp=valid_zoned_datetimes,
        value=int32s() | floats(allow_nan=False, allow_infinity=False),
    )
    async def test_main(
        self, *, data: DataObject, timestamp: dt.datetime, value: float
    ) -> None:
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis():
                    result = time_series_add(
                        container.ts, container.key, timestamp, value
                    )
                case redis.asyncio.Redis():
                    result = await time_series_add_async(
                        container.ts, container.key, timestamp, value
                    )
            assert isinstance(result, int)
            match container.client:
                case redis.Redis():
                    res_timestamp, res_value = time_series_get(
                        container.ts, container.key
                    )
                case redis.asyncio.Redis():
                    res_timestamp, res_value = await time_series_get_async(
                        container.ts, container.key
                    )
            assert res_timestamp == timestamp.astimezone(UTC)
            assert res_value == value

    @given(
        data=data(),
        timestamp=valid_zoned_datetimes,
        value=int32s() | floats(allow_nan=False, allow_infinity=False),
    )
    async def test_error_at_upsert(
        self, *, data: DataObject, timestamp: dt.datetime, value: float
    ) -> None:
        match = "Error at upsert under DUPLICATE_POLICY == 'BLOCK'; got .*"
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis():
                    _ = time_series_add(container.ts, container.key, timestamp, value)
                    with raises(_TimeSeriesAddErrorAtUpsertError, match=match):
                        _ = time_series_add(
                            container.ts, container.key, timestamp, value
                        )
                case redis.asyncio.Redis():
                    _ = await time_series_add_async(
                        container.ts, container.key, timestamp, value
                    )
                    with raises(_TimeSeriesAddErrorAtUpsertError, match=match):
                        _ = await time_series_add_async(
                            container.ts, container.key, timestamp, value
                        )

    @given(
        data=data(),
        timestamp=invalid_zoned_datetimes,
        value=int32s() | floats(allow_nan=False, allow_infinity=False),
    )
    async def test_error_invalid_timestamp(
        self, *, data: DataObject, timestamp: dt.datetime, value: float
    ) -> None:
        _ = assume(timestamp < EPOCH_UTC)
        match = "Timestamp must be at least the Epoch; got .*"
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis():
                    with raises(_TimeSeriesAddInvalidTimestampError, match=match):
                        _ = time_series_add(
                            container.ts, container.key, timestamp, value
                        )
                case redis.asyncio.Redis():
                    with raises(_TimeSeriesAddInvalidTimestampError, match=match):
                        _ = await time_series_add_async(
                            container.ts, container.key, timestamp, value
                        )

    @given(
        data=data(),
        timestamp=valid_zoned_datetimes,
        value=sampled_from([inf, -inf, nan]),
    )
    async def test_error_invalid_value(
        self, *, data: DataObject, timestamp: dt.datetime, value: float
    ) -> None:
        match = "Invalid value; got .*"
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis():
                    with raises(_TimeSeriesAddInvalidValueError, match=match):
                        _ = time_series_add(
                            container.ts, container.key, timestamp, value
                        )
                case redis.asyncio.Redis():
                    with raises(_TimeSeriesAddInvalidValueError, match=match):
                        _ = await time_series_add_async(
                            container.ts, container.key, timestamp, value
                        )


@SKIPIF_CI_AND_NOT_LINUX
class TestTimeSeriesAddAndReadDataFrame:
    @given(
        data=data(),
        key_timestamp_values=lists_fixed_length(text_ascii(), 4, unique=True).map(
            tuple
        ),
        strategy_dtype1=sampled_from([
            (int32s(), Int64),
            (floats(allow_nan=False, allow_infinity=False), Float64),
        ]),
        strategy_dtype2=sampled_from([
            (int32s(), Int64),
            (floats(allow_nan=False, allow_infinity=False), Float64),
        ]),
        time_zone=sampled_from([HongKong, UTC]),
        series_names=lists_fixed_length(text_ascii(), 2, unique=True).map(tuple),
    )
    async def test_main(
        self,
        *,
        data: DataObject,
        strategy_dtype1: tuple[SearchStrategy[Number], DataType],
        strategy_dtype2: tuple[SearchStrategy[Number], DataType],
        key_timestamp_values: tuple[str, str, str, str],
        series_names: tuple[str, str],
        time_zone: ZoneInfo,
    ) -> None:
        timestamp1, timestamp2 = data.draw(
            tuples(valid_zoned_datetimes, valid_zoned_datetimes)
        )
        strategy1, dtype1 = strategy_dtype1
        strategy2, dtype2 = strategy_dtype2
        value11, value21 = data.draw(tuples(strategy1, strategy1))
        value12, value22 = data.draw(tuples(strategy2, strategy2))
        key, timestamp, column1, column2 = key_timestamp_values
        columns = column1, column2
        schema = {
            key: Utf8,
            timestamp: zoned_datetime(time_zone=time_zone),
            column1: dtype1,
            column2: dtype2,
        }
        async with redis_cms(data) as container:
            key1, key2 = keys = cast(
                tuple[str, str], tuple(f"{container.key}_{id_}" for id_ in series_names)
            )
            df = DataFrame(
                [
                    (key1, timestamp1, value11, value12),
                    (key2, timestamp2, value21, value22),
                ],
                schema=schema,
                orient="row",
            )
            match container.client:
                case redis.Redis():
                    time_series_add_dataframe(
                        container.ts, df, key=key, timestamp=timestamp
                    )
                    result = time_series_read_dataframe(
                        container.ts,
                        keys,
                        columns,
                        output_key=key,
                        output_timestamp=timestamp,
                        output_time_zone=time_zone,
                    )
                case redis.asyncio.Redis():
                    await time_series_add_dataframe_async(
                        container.ts, df, key=key, timestamp=timestamp
                    )
                    result = await time_series_read_dataframe_async(
                        container.ts,
                        keys,
                        columns,
                        output_key=key,
                        output_timestamp=timestamp,
                        output_time_zone=time_zone,
                    )
            check_polars_dataframe(result, height=2, schema_list=schema)
        assert_frame_equal(result, df)

    @given(data=data())
    async def test_error_add_key_missing(self, *, data: DataObject) -> None:
        df = DataFrame()
        match = "DataFrame must have a 'key' column; got .*"
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis():
                    with raises(_TimeSeriesAddDataFrameKeyMissingError, match=match):
                        _ = time_series_add_dataframe(container.ts, df)
                case redis.asyncio.Redis():
                    with raises(_TimeSeriesAddDataFrameKeyMissingError, match=match):
                        _ = await time_series_add_dataframe_async(container.ts, df)

    @given(data=data())
    async def test_error_add_timestamp_missing(self, *, data: DataObject) -> None:
        df = DataFrame(schema={"key": Utf8})
        match = "DataFrame must have a 'timestamp' column; got .*"
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis():
                    with raises(
                        _TimeSeriesAddDataFrameTimestampMissingError, match=match
                    ):
                        _ = time_series_add_dataframe(container.ts, df)
                case redis.asyncio.Redis():
                    with raises(
                        _TimeSeriesAddDataFrameTimestampMissingError, match=match
                    ):
                        _ = await time_series_add_dataframe_async(container.ts, df)

    @given(data=data())
    async def test_error_add_key_is_not_utf8(self, *, data: DataObject) -> None:
        df = DataFrame(schema={"key": Boolean, "timestamp": DatetimeUTC})
        match = "The 'key' column must be Utf8; got Boolean"
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis():
                    with raises(_TimeSeriesAddDataFrameKeyIsNotUtf8Error, match=match):
                        _ = time_series_add_dataframe(container.ts, df)
                case redis.asyncio.Redis():
                    with raises(_TimeSeriesAddDataFrameKeyIsNotUtf8Error, match=match):
                        _ = await time_series_add_dataframe_async(container.ts, df)

    @given(data=data())
    async def test_error_add_timestamp_is_not_a_zoned_datetime(
        self, *, data: DataObject
    ) -> None:
        df = DataFrame(schema={"key": Utf8, "timestamp": Boolean})
        match = "The 'timestamp' column must be a zoned Datetime; got Boolean"
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis():
                    with raises(
                        _TimeSeriesAddDataFrameTimestampIsNotAZonedDatetimeError,
                        match=match,
                    ):
                        _ = time_series_add_dataframe(container.ts, df)
                case redis.asyncio.Redis():
                    with raises(
                        _TimeSeriesAddDataFrameTimestampIsNotAZonedDatetimeError,
                        match=match,
                    ):
                        _ = await time_series_add_dataframe_async(container.ts, df)

    @given(data=data())
    async def test_error_read_no_keys_requested(self, *, data: DataObject) -> None:
        match = "At least 1 key must be requested"
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis():
                    with raises(
                        _TimeSeriesReadDataFrameNoKeysRequestedError, match=match
                    ):
                        _ = time_series_read_dataframe(container.ts, [], [])
                case redis.asyncio.Redis():
                    with raises(
                        _TimeSeriesReadDataFrameNoKeysRequestedError, match=match
                    ):
                        _ = await time_series_read_dataframe_async(container.ts, [], [])

    @given(data=data())
    async def test_error_read_no_columns_requested(self, *, data: DataObject) -> None:
        match = "At least 1 column must be requested"
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis():
                    with raises(
                        _TimeSeriesReadDataFrameNoColumnsRequestedError, match=match
                    ):
                        _ = time_series_read_dataframe(container.ts, container.key, [])
                case redis.asyncio.Redis():
                    with raises(
                        _TimeSeriesReadDataFrameNoColumnsRequestedError, match=match
                    ):
                        _ = await time_series_read_dataframe_async(
                            container.ts, container.key, []
                        )


@SKIPIF_CI_AND_NOT_LINUX
class TestTimeSeriesMAddAndRange:
    int_schema: ClassVar[SchemaDict] = {
        "key": Utf8,
        "timestamp": DatetimeUTC,
        "value": Int64,
    }

    @given(
        data=data(),
        series_names=lists_fixed_length(text_ascii(), 2, unique=True).map(tuple),
        time_zone=sampled_from([HongKong, UTC]),
        key_timestamp_value=lists_fixed_length(text_ascii(), 3, unique=True).map(tuple),
        strategy_dtype=sampled_from([
            (int32s(), Int64),
            (floats(allow_nan=False, allow_infinity=False), Float64),
        ]),
    )
    async def test_main(
        self,
        *,
        data: DataObject,
        series_names: tuple[str, str],
        time_zone: ZoneInfo,
        key_timestamp_value: tuple[str, str, str],
        strategy_dtype: tuple[SearchStrategy[Number], DataType],
    ) -> None:
        timestamps = data.draw(tuples(valid_zoned_datetimes, valid_zoned_datetimes))
        strategy, dtype = strategy_dtype
        values = data.draw(tuples(strategy, strategy))
        key, timestamp, value = key_timestamp_value
        async with redis_cms(data) as container:
            keys = cast(
                tuple[str, str],
                tuple(f"{container.key}_{name}" for name in series_names),
            )
            triples = list(zip(keys, timestamps, values, strict=True))
            schema = {
                key: Utf8,
                timestamp: zoned_datetime(time_zone=time_zone),
                value: dtype,
            }
            values_or_df = data.draw(
                sampled_from([triples, DataFrame(triples, schema=schema, orient="row")])
            )
            match container.client:
                case redis.Redis():
                    res_madd = time_series_madd(
                        container.ts,
                        values_or_df,
                        key=key,
                        timestamp=timestamp,
                        value=value,
                    )
                case redis.asyncio.Redis():
                    res_madd = await time_series_madd_async(
                        container.ts,
                        values_or_df,
                        key=key,
                        timestamp=timestamp,
                        value=value,
                    )
            for i in res_madd:
                assert isinstance(i, int)
            match container.client:
                case redis.Redis():
                    res_range = time_series_range(
                        container.ts,
                        keys,
                        output_key=key,
                        output_timestamp=timestamp,
                        output_time_zone=time_zone,
                        output_value=value,
                    )
                case redis.asyncio.Redis():
                    res_range = await time_series_range_async(
                        container.ts,
                        keys,
                        output_key=key,
                        output_timestamp=timestamp,
                        output_time_zone=time_zone,
                        output_value=value,
                    )
            check_polars_dataframe(res_range, height=2, schema_list=schema)
            assert res_range.rows() == triples

    @given(data=data())
    async def test_error_madd_key_missing(self, *, data: DataObject) -> None:
        df = DataFrame()
        match = "DataFrame must have a 'key' column; got .*"
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis():
                    with raises(_TimeSeriesMAddKeyMissingError, match=match):
                        _ = time_series_madd(container.ts, df)
                case redis.asyncio.Redis():
                    with raises(_TimeSeriesMAddKeyMissingError, match=match):
                        _ = await time_series_madd_async(container.ts, df)

    @given(data=data())
    async def test_error_madd_timestamp_missing(self, *, data: DataObject) -> None:
        df = DataFrame(schema={"key": Utf8})
        match = "DataFrame must have a 'timestamp' column; got .*"
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis():
                    with raises(_TimeSeriesMAddTimestampMissingError, match=match):
                        _ = time_series_madd(container.ts, df)
                case redis.asyncio.Redis():
                    with raises(_TimeSeriesMAddTimestampMissingError, match=match):
                        _ = await time_series_madd_async(container.ts, df)

    @given(data=data())
    async def test_error_madd_value_missing(self, *, data: DataObject) -> None:
        df = DataFrame(schema={"key": Utf8, "timestamp": DatetimeUTC})
        match = "DataFrame must have a 'value' column; got .*"
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis():
                    with raises(_TimeSeriesMAddValueMissingError, match=match):
                        _ = time_series_madd(container.ts, df)
                case redis.asyncio.Redis():
                    with raises(_TimeSeriesMAddValueMissingError, match=match):
                        _ = await time_series_madd_async(container.ts, df)

    @given(data=data())
    async def test_error_madd_key_is_not_utf8(self, *, data: DataObject) -> None:
        df = DataFrame(
            schema={"key": Boolean, "timestamp": DatetimeUTC, "value": Float64}
        )
        match = "The 'key' column must be Utf8; got Boolean"
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis():
                    with raises(_TimeSeriesMAddKeyIsNotUtf8Error, match=match):
                        _ = time_series_madd(container.ts, df)
                case redis.asyncio.Redis():
                    with raises(_TimeSeriesMAddKeyIsNotUtf8Error, match=match):
                        _ = await time_series_madd_async(container.ts, df)

    @given(data=data())
    async def test_error_madd_timestamp_is_not_a_zoned_datetime(
        self, *, data: DataObject
    ) -> None:
        df = DataFrame(schema={"key": Utf8, "timestamp": Boolean, "value": Float64})
        match = "The 'timestamp' column must be a zoned Datetime; got Boolean"
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis():
                    with raises(
                        _TimeSeriesMAddTimestampIsNotAZonedDatetimeError, match=match
                    ):
                        _ = time_series_madd(container.ts, df)
                case redis.asyncio.Redis():
                    with raises(
                        _TimeSeriesMAddTimestampIsNotAZonedDatetimeError, match=match
                    ):
                        _ = await time_series_madd_async(container.ts, df)

    @given(data=data())
    async def test_error_madd_value_is_not_numeric(self, *, data: DataObject) -> None:
        df = DataFrame(schema={"key": Utf8, "timestamp": DatetimeUTC, "value": Boolean})
        match = "The 'value' column must be numeric; got Boolean"
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis():
                    with raises(_TimeSeriesMAddValueIsNotNumericError, match=match):
                        _ = time_series_madd(container.ts, df)
                case redis.asyncio.Redis():
                    with raises(_TimeSeriesMAddValueIsNotNumericError, match=match):
                        _ = await time_series_madd_async(container.ts, df)

    @given(data=data(), timestamp=valid_zoned_datetimes, value=int32s())
    async def test_error_madd_invalid_key(
        self, *, data: DataObject, timestamp: dt.datetime, value: int
    ) -> None:
        match = "The key '.*' must exist"
        async with redis_cms(data) as container:
            values = [(container.key, timestamp, value)]
            values_or_df = data.draw(
                sampled_from([
                    values,
                    DataFrame(values, schema=self.int_schema, orient="row"),
                ])
            )
            match container.client:
                case redis.Redis():
                    with raises(_TimeSeriesMAddInvalidKeyError, match=match):
                        _ = time_series_madd(
                            container.ts, values_or_df, assume_time_series_exist=True
                        )
                case redis.asyncio.Redis():
                    with raises(_TimeSeriesMAddInvalidKeyError, match=match):
                        _ = await time_series_madd_async(
                            container.ts, values_or_df, assume_time_series_exist=True
                        )

    @given(data=data(), timestamp=invalid_zoned_datetimes)
    async def test_error_madd_invalid_timestamp(
        self, *, data: DataObject, timestamp: dt.datetime
    ) -> None:
        value = data.draw(int32s())
        match = "Timestamps must be at least the Epoch; got .*"
        async with redis_cms(data) as container:
            values = [(container.key, timestamp, value)]
            values_or_df = data.draw(
                sampled_from([
                    values,
                    DataFrame(values, schema=self.int_schema, orient="row"),
                ])
            )
            match container.client:
                case redis.Redis():
                    with raises(_TimeSeriesMAddInvalidTimestampError, match=match):
                        _ = time_series_madd(container.ts, values_or_df)
                case redis.asyncio.Redis():
                    with raises(_TimeSeriesMAddInvalidTimestampError, match=match):
                        _ = await time_series_madd_async(container.ts, values_or_df)

    @given(
        data=data(),
        timestamp=valid_zoned_datetimes,
        value=sampled_from([inf, -inf, nan]),
    )
    async def test_error_madd_invalid_value(
        self, *, data: DataObject, timestamp: dt.datetime, value: float
    ) -> None:
        timestamp = data.draw(valid_zoned_datetimes)
        match = "The value .* is invalid"
        schema = {"key": Utf8, "timestamp": DatetimeUTC, "value": Float64}
        async with redis_cms(data) as container:
            values = [(f"{container.key}", timestamp, value)]
            values_or_df = data.draw(
                sampled_from([values, DataFrame(values, schema=schema, orient="row")])
            )
            match container.client:
                case redis.Redis():
                    with raises(_TimeSeriesMAddInvalidValueError, match=match):
                        _ = time_series_madd(container.ts, values_or_df)
                case redis.asyncio.Redis():
                    with raises(_TimeSeriesMAddInvalidValueError, match=match):
                        _ = await time_series_madd_async(container.ts, values_or_df)

    @given(data=data())
    async def test_error_range_no_keys_requested(self, *, data: DataObject) -> None:
        match = "At least 1 key must be requested; got .*"
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis():
                    with raises(_TimeSeriesRangeNoKeysRequestedError, match=match):
                        _ = time_series_range(container.ts, [])
                case redis.asyncio.Redis():
                    with raises(_TimeSeriesRangeNoKeysRequestedError, match=match):
                        _ = await time_series_range_async(container.ts, [])

    @given(data=data())
    async def test_error_range_invalid_key(self, *, data: DataObject) -> None:
        match = "The key '.*' must exist"
        async with redis_cms(data) as container:
            match container.client:
                case redis.Redis():
                    with raises(_TimeSeriesRangeInvalidKeyError, match=match):
                        _ = time_series_range(container.ts, container.key)
                case redis.asyncio.Redis():
                    with raises(_TimeSeriesRangeInvalidKeyError, match=match):
                        _ = await time_series_range_async(container.ts, container.key)

    @given(data=data(), timestamp=valid_zoned_datetimes, value=int32s())
    async def test_error_range_key_with_int64_and_float64(
        self, *, data: DataObject, timestamp: dt.datetime, value: int
    ) -> None:
        match = "The key '.*' contains both Int64 and Float64 data"
        async with redis_cms(data) as container:
            values = (
                [(container.key, timestamp, value)],
                [(container.key, timestamp, float(value))],
            )
            match container.client:
                case redis.Redis():
                    for vals in values:
                        _ = time_series_madd(container.ts, vals)
                    with raises(
                        _TimeSeriesRangeKeyWithInt64AndFloat64Error, match=match
                    ):
                        _ = time_series_range(container.ts, container.key)
                case redis.asyncio.Redis():
                    for vals in values:
                        _ = await time_series_madd_async(container.ts, vals)
                    with raises(
                        _TimeSeriesRangeKeyWithInt64AndFloat64Error, match=match
                    ):
                        _ = await time_series_range_async(container.ts, container.key)


class TestYieldClient:
    def test_sync(self) -> None:
        with yield_client() as client:
            assert isinstance(client, redis.Redis)

    async def test_async(self) -> None:
        async with yield_client_async() as client:
            assert isinstance(client, redis.asyncio.Redis)


class TestYieldTimeSeries:
    def test_sync(self) -> None:
        with yield_time_series() as ts:
            assert isinstance(ts, TimeSeries)

    async def test_async(self) -> None:
        async with yield_time_series_async() as ts:
            assert isinstance(ts, TimeSeries)



from __future__ import annotations

from operator import sub
from typing import cast

from hypothesis import given
from hypothesis.strategies import booleans, integers
from pytest import raises

from utilities.functions import EnsureNotNoneError, ensure_not_none
from utilities.functools import cache, lru_cache, partial


class TestCache:
    def test_main(self) -> None:
        counter = 0

        @cache
        def func(x: int, /) -> int:
            nonlocal counter
            counter += 1
            return x

        for _ in range(2):
            assert func(0) == 0
        assert counter == 1


class TestEnsureNotNone:
    def test_main(self) -> None:
        maybe_int = cast(int | None, 0)
        result = ensure_not_none(maybe_int)
        assert result == 0

    def test_error(self) -> None:
        with raises(EnsureNotNoneError, match="Object .* must not be None"):
            _ = ensure_not_none(None)


class TestLRUCache:
    def test_no_arguments(self) -> None:
        counter = 0

        @lru_cache
        def func(x: int, /) -> int:
            nonlocal counter
            counter += 1
            return x

        for _ in range(2):
            assert func(0) == 0
        assert counter == 1

    @given(max_size=integers(1, 10), typed=booleans())
    def test_with_arguments(self, *, max_size: int, typed: bool) -> None:
        counter = 0

        @lru_cache(max_size=max_size, typed=typed)
        def func(x: int, /) -> int:
            nonlocal counter
            counter += 1
            return x

        for _ in range(2):
            assert func(0) == 0
        assert counter == 1


class TestPartial:
    @given(x=integers(), y=integers())
    def test_main(self, *, x: int, y: int) -> None:
        func = partial(sub, ..., y)
        assert func(x) == x - y



from __future__ import annotations

import sys
from functools import wraps
from operator import neg
from types import NoneType
from typing import TYPE_CHECKING, Any, TypeVar

from hypothesis import given
from hypothesis.strategies import booleans, integers
from pytest import CaptureFixture, mark, param, raises

from utilities.asyncio import try_await
from utilities.functions import (
    ensure_not_none,
    first,
    get_class,
    get_class_name,
    get_func_name,
    identity,
    is_none,
    is_not_none,
    not_func,
    second,
    send_and_next,
    start_generator_coroutine,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Generator

_T = TypeVar("_T")


class TestFirst:
    @given(x=integers(), y=integers())
    def test_main(self, *, x: int, y: int) -> None:
        pair = x, y
        assert first(pair) == x


class TestGetClass:
    @mark.parametrize(
        ("obj", "expected"), [param(None, NoneType), param(NoneType, NoneType)]
    )
    def test_main(self, *, obj: Any, expected: type[Any]) -> None:
        assert get_class(obj) is expected


class TestGetClassName:
    def test_class(self) -> None:
        class Example: ...

        assert get_class_name(Example) == "Example"

    def test_instance(self) -> None:
        class Example: ...

        assert get_class_name(Example()) == "Example"


class TestGetFuncName:
    @mark.parametrize(
        ("func", "expected"),
        [
            param(identity, "identity"),
            param(lambda x: x, "<lambda>"),  # pyright: ignore[reportUnknownLambdaType]
            param(len, "len"),
            param(neg, "neg"),
            param(object.__init__, "object.__init__"),
            param(object().__str__, "object.__str__"),
            param(repr, "repr"),
            param(str, "str"),
            param(try_await, "try_await"),
            param(str.join, "str.join"),
            param(sys.exit, "exit"),
        ],
    )
    def test_main(self, *, func: Callable[..., Any], expected: str) -> None:
        assert get_func_name(func) == expected

    def test_decorated(self) -> None:
        @wraps(identity)
        def wrapped(x: _T, /) -> _T:
            return identity(x)

        assert get_func_name(wrapped) == "identity"

    def test_object(self) -> None:
        class Example:
            def __call__(self, x: _T, /) -> _T:
                return identity(x)

        obj = Example()
        assert get_func_name(obj) == "Example"

    def test_obj_method(self) -> None:
        class Example:
            def obj_method(self, x: _T) -> _T:
                return identity(x)

        obj = Example()
        assert get_func_name(obj.obj_method) == "obj_method"

    def test_obj_classmethod(self) -> None:
        class Example:
            @classmethod
            def obj_classmethod(cls: _T) -> _T:
                return identity(cls)

        assert get_func_name(Example.obj_classmethod) == "obj_classmethod"

    def test_obj_staticmethod(self) -> None:
        class Example:
            @staticmethod
            def obj_staticmethod(x: _T) -> _T:
                return identity(x)

        assert get_func_name(Example.obj_staticmethod) == "obj_staticmethod"


class TestIdentity:
    @given(x=integers())
    def test_main(self, *, x: int) -> None:
        assert identity(x) == x


class TestIsNoneAndIsNotNone:
    @mark.parametrize(
        ("func", "obj", "expected"),
        [
            param(is_none, None, True),
            param(is_none, 0, False),
            param(is_not_none, None, False),
            param(is_not_none, 0, True),
        ],
    )
    def test_main(
        self, *, func: Callable[[Any], bool], obj: Any, expected: bool
    ) -> None:
        result = func(obj)
        assert result is expected


class TestNotFunc:
    @given(x=booleans())
    def test_main(self, *, x: bool) -> None:
        def return_x() -> bool:
            return x

        return_not_x = not_func(return_x)
        result = return_not_x()
        expected = not x
        assert result is expected


class TestSecond:
    @given(x=integers(), y=integers())
    def test_main(self, *, x: int, y: int) -> None:
        pair = x, y
        assert second(pair) == y


class TestSendAndNext:
    def test_main(self, *, capsys: CaptureFixture) -> None:
        @start_generator_coroutine
        def func() -> Generator[int | None, float | None, str]:
            print("Initial")  # noqa: T201
            while True:
                input_ = yield
                output = round(ensure_not_none(input_))
                if output >= 0:
                    print(f"Received {input_}, yielding {output}")  # noqa: T201
                    yield output
                else:
                    return "Done"

        generator = func()
        out = capsys.readouterr().out
        assert out == "Initial\n", out
        result = send_and_next(0.1, generator)
        assert result == 0
        out = capsys.readouterr().out
        assert out == "Received 0.1, yielding 0\n", out
        result = send_and_next(0.9, generator)
        assert result == 1
        out = capsys.readouterr().out
        assert out == "Received 0.9, yielding 1\n", out
        result = send_and_next(1.1, generator)
        assert result == 1
        out = capsys.readouterr().out
        assert out == "Received 1.1, yielding 1\n", out
        with raises(StopIteration) as exc:
            _ = send_and_next(-0.9, generator)
        assert exc.value.args == ("Done",)


class TestStartGeneratorCoroutine:
    def test_main(self, *, capsys: CaptureFixture) -> None:
        @start_generator_coroutine
        def func() -> Generator[int, float, str]:
            print("Pre-initial")  # noqa: T201
            x = yield 0
            print(f"Post-initial; x={x}")  # noqa: T201
            while x >= 0:
                print(f"Pre-yield; x={x}")  # noqa: T201
                x = yield round(x)
                print(f"Post-yield; x={x}")  # noqa: T201
            return "Done"

        generator = func()
        out = capsys.readouterr().out
        assert out == "Pre-initial\n", out
        assert generator.send(0.1) == 0
        out = capsys.readouterr().out
        assert out == "Post-initial; x=0.1\nPre-yield; x=0.1\n", out
        assert generator.send(0.9) == 1
        out = capsys.readouterr().out
        assert out == "Post-yield; x=0.9\nPre-yield; x=0.9\n", out
        assert generator.send(1.1) == 1
        out = capsys.readouterr().out
        assert out == "Post-yield; x=1.1\nPre-yield; x=1.1\n", out
        with raises(StopIteration) as exc:
            _ = generator.send(-0.9)
        assert exc.value.args == ("Done",)



from __future__ import annotations

import datetime as dt
from re import escape
from zoneinfo import ZoneInfo

from hypothesis import given
from hypothesis.strategies import DataObject, data, sampled_from
from pytest import mark, param, raises

from utilities.zoneinfo import (
    UTC,
    EnsureTimeZoneError,
    HongKong,
    Tokyo,
    USCentral,
    USEastern,
    ensure_time_zone,
    get_time_zone_name,
)


class TestGetTimeZoneName:
    @given(data=data())
    @mark.parametrize(
        "time_zone",
        [
            param("Asia/Hong_Kong"),
            param("Asia/Tokyo"),
            param("US/Central"),
            param("US/Eastern"),
            param("UTC"),
        ],
    )
    def test_main(self, *, data: DataObject, time_zone: str) -> None:
        zone_info_or_str = data.draw(sampled_from([ZoneInfo(time_zone), time_zone]))
        result = get_time_zone_name(zone_info_or_str)
        assert result == time_zone


class TestEnsureZoneInfo:
    @given(data=data())
    @mark.parametrize(
        ("time_zone", "expected"),
        [
            param(HongKong, HongKong),
            param(Tokyo, Tokyo),
            param(USCentral, USCentral),
            param(USEastern, USEastern),
            param(UTC, UTC),
            param(dt.UTC, UTC),
        ],
    )
    def test_main(
        self, *, data: DataObject, time_zone: ZoneInfo | dt.timezone, expected: ZoneInfo
    ) -> None:
        zone_info_or_str = data.draw(
            sampled_from([time_zone, get_time_zone_name(time_zone)])
        )
        result = ensure_time_zone(zone_info_or_str)
        assert result is expected

    def test_error(self) -> None:
        time_zone = dt.timezone(dt.timedelta(hours=12))
        with raises(
            EnsureTimeZoneError, match=escape("Unsupported time zone: UTC+12:00")
        ):
            _ = ensure_time_zone(time_zone)


class TestTimeZones:
    @mark.parametrize(
        "time_zone", [param(HongKong), param(Tokyo), param(USCentral), param(USEastern)]
    )
    def test_main(self, *, time_zone: ZoneInfo) -> None:
        assert isinstance(time_zone, ZoneInfo)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Callable

    from eventkit import Event


def add_listener(
    event: Event,
    listener: Callable[..., Any],
    /,
    *,
    error: Callable[..., Any] | None = None,
    done: Callable[..., Any] | None = None,
    keep_ref: bool = False,
) -> Event:
    """Connect a listener to an event."""
    return event.connect(listener, error=error, done=done, keep_ref=keep_ref)


__all__ = ["add_listener"]



from __future__ import annotations

import datetime as dt
import decimal
from contextlib import suppress
from dataclasses import dataclass
from itertools import chain
from typing import TYPE_CHECKING, Any, Literal, cast, overload
from uuid import UUID

import polars as pl
from polars import (
    Binary,
    DataFrame,
    Date,
    Datetime,
    Duration,
    Float64,
    Int32,
    Int64,
    Time,
    UInt32,
    UInt64,
    Utf8,
    concat,
    read_database,
)
from polars._typing import ConnectionOrCursor, PolarsDataType, SchemaDict
from sqlalchemy import Column, Select, Table, select
from sqlalchemy.exc import DuplicateColumnError
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine
from typing_extensions import override

from utilities.datetime import is_subclass_date_not_datetime
from utilities.errors import redirect_error
from utilities.functions import ensure_not_none, identity
from utilities.iterables import (
    CheckDuplicatesError,
    OneError,
    check_duplicates,
    chunked,
    one,
)
from utilities.polars import (
    EmptyPolarsConcatError,
    redirect_empty_polars_concat,
    zoned_datetime,
)
from utilities.sqlalchemy import (
    CHUNK_SIZE_FRAC,
    EngineOrConnection,
    MaybeAsyncEngineOrConnection,
    TableOrMappedClass,
    ensure_tables_created,
    ensure_tables_created_async,
    get_chunk_size,
    get_columns,
    insert_items,
    insert_items_async,
    upsert_items,
    upsert_items_async,
    yield_connection,
)
from utilities.zoneinfo import UTC

if TYPE_CHECKING:
    from collections.abc import (
        AsyncIterable,
        AsyncIterator,
        Iterable,
        Iterator,
        Mapping,
        Sequence,
    )
    from zoneinfo import ZoneInfo

    from sqlalchemy.sql import ColumnCollection
    from sqlalchemy.sql.base import ReadOnlyColumnCollection

    from utilities.types import StrMapping


def insert_dataframe(
    df: DataFrame,
    table_or_mapped_class: TableOrMappedClass,
    engine_or_conn: EngineOrConnection,
    /,
    *,
    snake: bool = False,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    assume_tables_exist: bool = False,
) -> None:
    """Insert a DataFrame into a database."""
    prepared = _insert_dataframe_prepare(df, table_or_mapped_class, snake=snake)
    if prepared.no_items_empty_df:
        if not assume_tables_exist:
            ensure_tables_created(engine_or_conn, table_or_mapped_class)
        return
    if prepared.no_items_non_empty_df:
        raise InsertDataFrameError(df=df)
    insert_items(
        engine_or_conn,
        prepared.items,
        chunk_size_frac=chunk_size_frac,
        assume_tables_exist=assume_tables_exist,
    )


@dataclass(kw_only=True)
class InsertDataFrameError(Exception):
    df: DataFrame

    @override
    def __str__(self) -> str:
        return f"Non-empty DataFrame must resolve to at least 1 item\n\n{self.df}"


async def insert_dataframe_async(
    df: DataFrame,
    table_or_mapped_class: TableOrMappedClass,
    engine: AsyncEngine,  # does not support connection
    /,
    *,
    snake: bool = False,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    assume_tables_exist: bool = False,
) -> None:
    """Insert a DataFrame into a database."""
    prepared = _insert_dataframe_prepare(df, table_or_mapped_class, snake=snake)
    if prepared.no_items_empty_df:
        if not assume_tables_exist:
            await ensure_tables_created_async(engine, table_or_mapped_class)
        return
    if prepared.no_items_non_empty_df:
        raise InsertDataFrameAsyncError(df=df)
    await insert_items_async(
        engine,
        prepared.items,
        chunk_size_frac=chunk_size_frac,
        assume_tables_exist=assume_tables_exist,
    )


@dataclass(kw_only=True)
class InsertDataFrameAsyncError(Exception):
    df: DataFrame

    @override
    def __str__(self) -> str:
        return f"Non-empty DataFrame must resolve to at least 1 item\n\n{self.df}"


@dataclass(frozen=True, kw_only=True)
class _InsertDataFramePrepare:
    items: tuple[Sequence[StrMapping], TableOrMappedClass]
    no_items_empty_df: bool
    no_items_non_empty_df: bool


def _insert_dataframe_prepare(
    df: DataFrame, table_or_mapped_class: Table | type[Any], /, *, snake: bool = False
) -> _InsertDataFramePrepare:
    """Prepare the arguments for `insert_dataframe`."""
    mapping = _insert_dataframe_map_df_schema_to_table(
        df.schema, table_or_mapped_class, snake=snake
    )
    items = df.select(mapping).rename(mapping).to_dicts()
    no_items = len(items) == 0
    df_is_empty = df.is_empty()
    return _InsertDataFramePrepare(
        items=(items, table_or_mapped_class),
        no_items_empty_df=no_items and df_is_empty,
        no_items_non_empty_df=no_items and not df_is_empty,
    )


def _insert_dataframe_map_df_schema_to_table(
    df_schema: SchemaDict,
    table_or_mapped_class: TableOrMappedClass,
    /,
    *,
    snake: bool = False,
) -> dict[str, str]:
    """Map a DataFrame schema to a table."""
    table_schema = {
        col.name: col.type.python_type for col in get_columns(table_or_mapped_class)
    }
    out: dict[str, str] = {}
    for df_col_name, df_col_type in df_schema.items():
        with suppress(_InsertDataFrameMapDFColumnToTableColumnAndTypeError):
            out[df_col_name] = _insert_dataframe_map_df_column_to_table_schema(
                df_col_name, df_col_type, table_schema, snake=snake
            )
    return out


def _insert_dataframe_map_df_column_to_table_schema(
    df_col_name: str,
    df_col_type: PolarsDataType,
    table_schema: Mapping[str, type],
    /,
    *,
    snake: bool = False,
) -> str:
    """Map a DataFrame column to a table schema."""
    db_col_name, db_col_type = _insert_dataframe_map_df_column_to_table_column_and_type(
        df_col_name, table_schema, snake=snake
    )
    if not _insert_dataframe_check_df_and_db_types(df_col_type, db_col_type):
        msg = f"{df_col_type=}, {db_col_type=}"
        raise _InsertDataFrameMapDFColumnToTableSchemaError(msg)
    return db_col_name


class _InsertDataFrameMapDFColumnToTableSchemaError(Exception): ...


def _insert_dataframe_map_df_column_to_table_column_and_type(
    df_col_name: str, table_schema: Mapping[str, type], /, *, snake: bool = False
) -> tuple[str, type]:
    """Map a DataFrame column to a table column and type."""
    from utilities.humps import snake_case

    items = table_schema.items()
    func = snake_case if snake else identity
    target = func(df_col_name)
    with redirect_error(
        OneError,
        _InsertDataFrameMapDFColumnToTableColumnAndTypeError(
            f"{df_col_name=}, {table_schema=}, {snake=}"
        ),
    ):
        return one((n, t) for n, t in items if func(n) == target)


class _InsertDataFrameMapDFColumnToTableColumnAndTypeError(Exception): ...


def _insert_dataframe_check_df_and_db_types(
    dtype: PolarsDataType, db_col_type: type, /
) -> bool:
    return (
        (dtype == pl.Boolean and issubclass(db_col_type, bool))
        or (dtype == Date and is_subclass_date_not_datetime(db_col_type))
        or (dtype == Datetime and issubclass(db_col_type, dt.datetime))
        or (dtype == Float64 and issubclass(db_col_type, float))
        or (dtype in {Int32, Int64, UInt32, UInt64} and issubclass(db_col_type, int))
        or (dtype == Utf8 and issubclass(db_col_type, str))
    )


@overload
def select_to_dataframe(
    sel: Select[Any],
    engine_or_conn: EngineOrConnection,
    /,
    *,
    snake: bool = ...,
    time_zone: ZoneInfo | str = ...,
    batch_size: None = ...,
    in_clauses: tuple[Column[Any], Iterable[Any]] | None = ...,
    in_clauses_chunk_size: int | None = ...,
    chunk_size_frac: float = ...,
    **kwargs: Any,
) -> DataFrame: ...
@overload
def select_to_dataframe(
    sel: Select[Any],
    engine_or_conn: EngineOrConnection,
    /,
    *,
    snake: bool = ...,
    time_zone: ZoneInfo | str = ...,
    batch_size: int = ...,
    in_clauses: tuple[Column[Any], Iterable[Any]] | None = ...,
    in_clauses_chunk_size: int | None = ...,
    chunk_size_frac: float = ...,
    **kwargs: Any,
) -> Iterable[DataFrame]: ...
def select_to_dataframe(
    sel: Select[Any],
    engine_or_conn: EngineOrConnection,
    /,
    *,
    snake: bool = False,
    time_zone: ZoneInfo | str = UTC,
    batch_size: int | None = None,
    in_clauses: tuple[Column[Any], Iterable[Any]] | None = None,
    in_clauses_chunk_size: int | None = None,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    **kwargs: Any,
) -> DataFrame | Iterable[DataFrame]:
    """Read a table from a database into a DataFrame."""
    prepared = _select_to_dataframe_prepare(
        sel,
        engine_or_conn,
        snake=snake,
        time_zone=time_zone,
        in_clauses=in_clauses,
        in_clauses_chunk_size=in_clauses_chunk_size,
        chunk_size_frac=chunk_size_frac,
    )
    if in_clauses is None:
        return read_database(
            prepared.sel,
            cast(ConnectionOrCursor, engine_or_conn),
            iter_batches=batch_size is not None,
            batch_size=batch_size,
            schema_overrides=prepared.schema,
            **kwargs,
        )
    sels = ensure_not_none(prepared.sels)
    if batch_size is None:
        with yield_connection(engine_or_conn) as conn:
            dfs = (
                select_to_dataframe(
                    sel,
                    conn,
                    snake=snake,
                    time_zone=time_zone,
                    batch_size=None,
                    in_clauses=None,
                    **kwargs,
                )
                for sel in sels
            )
            try:
                with redirect_empty_polars_concat():
                    return concat(dfs)
            except EmptyPolarsConcatError:
                return DataFrame(schema=prepared.schema)
    dfs = (
        select_to_dataframe(
            sel,
            engine_or_conn,
            snake=snake,
            time_zone=time_zone,
            batch_size=batch_size,
            in_clauses=None,
            chunk_size_frac=chunk_size_frac,
            **kwargs,
        )
        for sel in sels
    )
    return chain(*dfs)


@overload
async def select_to_dataframe_async(
    sel: Select[Any],
    engine: AsyncEngine,
    /,
    *,
    snake: bool = ...,
    time_zone: ZoneInfo | str = ...,
    batch_size: None = ...,
    in_clauses: tuple[Column[Any], Iterable[Any]] | None = ...,
    in_clauses_chunk_size: int | None = ...,
    chunk_size_frac: float = ...,
    **kwargs: Any,
) -> DataFrame: ...
@overload
async def select_to_dataframe_async(
    sel: Select[Any],
    engine: AsyncEngine,
    /,
    *,
    snake: bool = ...,
    time_zone: ZoneInfo | str = ...,
    batch_size: int = ...,
    in_clauses: None = ...,
    in_clauses_chunk_size: int | None = ...,
    chunk_size_frac: float = ...,
    **kwargs: Any,
) -> Iterable[DataFrame]: ...
@overload
async def select_to_dataframe_async(
    sel: Select[Any],
    engine: AsyncEngine,
    /,
    *,
    snake: bool = ...,
    time_zone: ZoneInfo | str = ...,
    batch_size: int = ...,
    in_clauses: tuple[Column[Any], Iterable[Any]] = ...,
    in_clauses_chunk_size: int | None = ...,
    chunk_size_frac: float = ...,
    **kwargs: Any,
) -> AsyncIterable[DataFrame]: ...
async def select_to_dataframe_async(
    sel: Select[Any],
    engine: AsyncEngine,
    /,
    *,
    snake: bool = False,
    time_zone: ZoneInfo | str = UTC,
    batch_size: int | None = None,
    in_clauses: tuple[Column[Any], Iterable[Any]] | None = None,
    in_clauses_chunk_size: int | None = None,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    **kwargs: Any,
) -> DataFrame | Iterable[DataFrame] | AsyncIterable[DataFrame]:
    """Read a table from a database into a DataFrame."""
    if not issubclass(AsyncEngine, type(engine)):
        # for handling testing
        engine = create_async_engine(engine.url)
        return await select_to_dataframe_async(
            sel,
            engine,
            snake=snake,
            time_zone=time_zone,
            batch_size=batch_size,
            in_clauses=in_clauses,
            in_clauses_chunk_size=in_clauses_chunk_size,
            chunk_size_frac=chunk_size_frac,
            **kwargs,
        )
    prepared = _select_to_dataframe_prepare(
        sel,
        engine,
        snake=snake,
        time_zone=time_zone,
        in_clauses=in_clauses,
        in_clauses_chunk_size=in_clauses_chunk_size,
        chunk_size_frac=chunk_size_frac,
    )
    if in_clauses is None:
        return read_database(
            prepared.sel,
            cast(Any, engine),
            iter_batches=batch_size is not None,
            batch_size=batch_size,
            schema_overrides=prepared.schema,
            **kwargs,
        )
    sels = ensure_not_none(prepared.sels)
    if batch_size is None:
        dfs = [
            await select_to_dataframe_async(
                sel,
                engine,
                snake=snake,
                time_zone=time_zone,
                batch_size=None,
                in_clauses=None,
                **kwargs,
            )
            for sel in sels
        ]
        try:
            with redirect_empty_polars_concat():
                return concat(dfs)
        except EmptyPolarsConcatError:
            return DataFrame(schema=prepared.schema)

    async def yield_dfs() -> AsyncIterator[DataFrame]:
        for sel_i in sels:
            for df in await select_to_dataframe_async(
                sel_i,
                engine,
                snake=snake,
                time_zone=time_zone,
                batch_size=batch_size,
                in_clauses=None,
                chunk_size_frac=chunk_size_frac,
                **kwargs,
            ):
                yield df

    return yield_dfs()


@dataclass(frozen=True, kw_only=True)
class _SelectToDataFramePrepare:
    sel: Select[Any]
    schema: SchemaDict
    sels: Iterable[Select[Any]] | None = None


def _select_to_dataframe_prepare(
    sel: Select[Any],
    engine_or_conn: EngineOrConnection | AsyncEngine,
    /,
    *,
    snake: bool = False,
    time_zone: ZoneInfo | str = UTC,
    in_clauses: tuple[Column[Any], Iterable[Any]] | None = None,
    in_clauses_chunk_size: int | None = None,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
) -> _SelectToDataFramePrepare:
    if snake:
        sel = _select_to_dataframe_apply_snake(sel)
    schema = _select_to_dataframe_map_select_to_df_schema(sel, time_zone=time_zone)
    if in_clauses is None:
        sels = None
    else:
        sels = _select_to_dataframe_yield_selects_with_in_clauses(
            sel,
            engine_or_conn,
            in_clauses,
            in_clauses_chunk_size=in_clauses_chunk_size,
            chunk_size_frac=chunk_size_frac,
        )
    return _SelectToDataFramePrepare(sel=sel, schema=schema, sels=sels)


def _select_to_dataframe_apply_snake(sel: Select[Any], /) -> Select[Any]:
    """Apply snake-case to a selectable."""
    from utilities.humps import snake_case

    alias = sel.alias()
    columns = [alias.c[c.name].label(snake_case(c.name)) for c in sel.selected_columns]
    return select(*columns)


def _select_to_dataframe_map_select_to_df_schema(
    sel: Select[Any], /, *, time_zone: ZoneInfo | str = UTC
) -> SchemaDict:
    """Map a select to a DataFrame schema."""
    columns: ReadOnlyColumnCollection = cast(Any, sel).selected_columns
    _select_to_dataframe_check_duplicates(columns)
    return {
        col.name: _select_to_dataframe_map_table_column_type_to_dtype(
            col.type, time_zone=time_zone
        )
        for col in columns
    }


def _select_to_dataframe_map_table_column_type_to_dtype(
    type_: Any, /, *, time_zone: ZoneInfo | str = UTC
) -> PolarsDataType:
    """Map a table column type to a polars type."""
    type_use = type_() if isinstance(type_, type) else type_
    py_type = type_use.python_type
    if issubclass(py_type, bool):
        return pl.Boolean
    if issubclass(py_type, bytes):
        return Binary
    if issubclass(py_type, decimal.Decimal):
        return pl.Decimal
    if issubclass(py_type, dt.date) and not issubclass(py_type, dt.datetime):
        return pl.Date
    if issubclass(py_type, dt.datetime):
        has_tz: bool = type_use.timezone
        return zoned_datetime(time_zone=time_zone) if has_tz else Datetime()
    if issubclass(py_type, dt.time):
        return Time
    if issubclass(py_type, dt.timedelta):
        return Duration
    if issubclass(py_type, float):
        return Float64
    if issubclass(py_type, int):
        return Int64
    if issubclass(py_type, UUID | str):
        return Utf8
    msg = f"{type_=}, {py_type=}"  # pragma: no cover
    raise _SelectToDataFrameMapTableColumnToDTypeError(msg)  # pragma: no cover


class _SelectToDataFrameMapTableColumnToDTypeError(Exception): ...


def _select_to_dataframe_check_duplicates(
    columns: ColumnCollection[Any, Any], /
) -> None:
    """Check a select for duplicate columns."""
    names = [col.name for col in columns]
    with redirect_error(CheckDuplicatesError, DuplicateColumnError(f"{names=}")):
        check_duplicates(names)


def _select_to_dataframe_yield_selects_with_in_clauses(
    sel: Select[Any],
    engine_or_conn: MaybeAsyncEngineOrConnection,
    in_clauses: tuple[Column[Any], Iterable[Any]],
    /,
    *,
    in_clauses_chunk_size: int | None = None,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
) -> Iterator[Select[Any]]:
    max_length = len(sel.selected_columns)
    in_col, in_values = in_clauses
    if in_clauses_chunk_size is None:
        chunk_size = get_chunk_size(
            engine_or_conn, chunk_size_frac=chunk_size_frac, scaling=max_length
        )
    else:
        chunk_size = in_clauses_chunk_size
    return (sel.where(in_col.in_(values)) for values in chunked(in_values, chunk_size))


def upsert_dataframe(
    df: DataFrame,
    table_or_mapped_class: TableOrMappedClass,
    engine_or_conn: EngineOrConnection,
    /,
    *,
    snake: bool = False,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    assume_tables_exist: bool = False,
    selected_or_all: Literal["selected", "all"] = "selected",
) -> None:
    """Upsert a DataFrame into a database."""
    prepared = _insert_dataframe_prepare(df, table_or_mapped_class, snake=snake)
    if prepared.no_items_empty_df:
        if not assume_tables_exist:
            ensure_tables_created(engine_or_conn, table_or_mapped_class)
        return
    if prepared.no_items_non_empty_df:
        raise UpsertDataFrameError(df=df)
    upsert_items(
        engine_or_conn,
        prepared.items,
        chunk_size_frac=chunk_size_frac,
        assume_tables_exist=assume_tables_exist,
        selected_or_all=selected_or_all,
    )


@dataclass(kw_only=True)
class UpsertDataFrameError(Exception):
    df: DataFrame

    @override
    def __str__(self) -> str:
        return f"Non-empty DataFrame must resolve to at least 1 item\n\n{self.df}"


async def upsert_dataframe_async(
    df: DataFrame,
    table_or_mapped_class: TableOrMappedClass,
    engine: AsyncEngine,  # does not support connection
    /,
    *,
    snake: bool = False,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    assume_tables_exist: bool = False,
    selected_or_all: Literal["selected", "all"] = "selected",
) -> None:
    """Upsert a DataFrame into a database."""
    prepared = _insert_dataframe_prepare(df, table_or_mapped_class, snake=snake)
    if prepared.no_items_empty_df:
        if not assume_tables_exist:
            await ensure_tables_created_async(engine, table_or_mapped_class)
        return
    if prepared.no_items_non_empty_df:
        raise UpsertDataFrameAsyncError(df=df)
    await upsert_items_async(
        engine,
        prepared.items,
        chunk_size_frac=chunk_size_frac,
        assume_tables_exist=assume_tables_exist,
        selected_or_all=selected_or_all,
    )


@dataclass(kw_only=True)
class UpsertDataFrameAsyncError(Exception):
    df: DataFrame

    @override
    def __str__(self) -> str:
        return f"Non-empty DataFrame must resolve to at least 1 item\n\n{self.df}"


__all__ = [
    "InsertDataFrameAsyncError",
    "InsertDataFrameError",
    "UpsertDataFrameAsyncError",
    "UpsertDataFrameError",
    "insert_dataframe",
    "insert_dataframe_async",
    "select_to_dataframe",
    "select_to_dataframe_async",
    "upsert_dataframe",
    "upsert_dataframe_async",
]
